/*
 * Copyright (C) 2025 IOPSYS Software Solutions AB. All rights reserved.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <pthread.h>
#include <inttypes.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <uci.h>

#include <easy/easy.h>

#include <cmdu.h>
#include <i1905_wsc.h>
#ifdef CONTROLLER_SYNC_DYNAMIC_CNTLR_CONFIG
#include <cntlrsync.h>
#endif
#include <1905_tlvs.h>
#include <easymesh.h>
#include <map_module.h>

#include <wifidefs.h>
#include "wifi_dataelements.h"

#include "timer.h"
#if (EASYMESH_VERSION > 2)
#ifdef USE_LIBDPP
#include <dpp_api.h>
#endif
#include "dpp.h"
#endif

#include "utils/utils.h"
#include "utils/debug.h"
#include "utils/liblist.h"
#include "backhaul_topology.h"
#include "config.h"
#include "sta.h"
#include "steer_module.h"
#include "steer.h"
#include "cntlr.h"
#include "scan.h"
#include "cntlr_ubus.h"

#include "cntlr_map_debug.h"
#include "cntlr_cmdu.h"
#include "cntlr_tlv.h"
#include "wifi_opclass.h"
#include "cntlr_map.h"


#if 0 // unused
/* use unassociated STA measurements to steer */
void cntlr_check_usta_steer(struct controller *c, struct sta *s)
{
	struct unassoc_sta_metrics *best = NULL, *u = NULL;
	struct node *n = s->fh->agent;
	struct netif_iface *best_fh = s->fh;
	struct steer_control_config *scc;

	scc  = get_steer_control_config(c);
	if (!scc)
		return;

	if (!scc->use_usta_metrics) {
		dbg("%s %d Will not use unassociated STA metrics \
		    data to steer\n", __func__, __LINE__);
		return;
	}

	dbg("%s %d for "MACFMT" attached to bssid " MACFMT " node = " \
	    MACFMT "\n", __func__, __LINE__, MAC2STR(s->de_sta->macaddr),
	    MAC2STR(s->bssid), MAC2STR(n->almacaddr));

	list_for_each_entry(u, &s->umetriclist, list) {
		dbg("%s %d check usta node "MACFMT"\n",
		    __func__, __LINE__, MAC2STR(u->agent->almacaddr));

		if (!best) {
			best = u;
			continue;
		}

		dbg("%s %d best ul_rcpi %u this ul_rcpi %u\n", __func__, __LINE__,
				best->ul_rcpi, u->ul_rcpi);

		if ((best->ul_rcpi - u->ul_rcpi) > USTA_STEER_UL_RCPI_DELTA) {
			dbg("%s %d new best usta node "MACFMT" with ul_rcpi %d\n",
					__func__, __LINE__,
					MAC2STR(u->agent->almacaddr),
					u->ul_rcpi);
			best = u;
		}
	}

	if (!best)
		return;

	if (best_fh && !hwaddr_is_zero(best_fh->bss->bssid)
			&& memcmp(best_fh->bss->bssid, s->bssid, 6)) {

		struct cmdu_buff *cmdu;
		int ret = 0;

		if ((s->type == IEEE1905 && !scc->enable_bsta_steer) ||
		    (s->type == NON_IEEE1905 && !scc->enable_sta_steer)) {
			trace("|%s:%d| better bssid found, but will not steer "MACFMT",\
			       because the 'enable_(b)sta_steer' is not set!\n",
			       __func__, __LINE__, MAC2STR(s->de_sta->macaddr));
			return;
		}

		dbg("%s %d better bssid found! try to steer " MACFMT " \
		    from " MACFMT " to " MACFMT "\n",
		    __func__, __LINE__,
		    MAC2STR(s->de_sta->macaddr), MAC2STR(s->bssid),
		    MAC2STR(best_fh->bss->bssid));

		if (s->type == IEEE1905) {
			cmdu = cntlr_gen_backhaul_steer_request(c, s->agent->almacaddr,
					s->de_sta->macaddr, best_fh->bss->bssid, 0, 0);
			if (cmdu) {
				send_cmdu(c, cmdu);
				cmdu_free(cmdu);
			}
		} else {
			ret = cntlr_send_client_steer_request(c, s->fh->agent->almacaddr,
					s->bssid, 0,
					1, (uint8_t (*)[6])s->de_sta->macaddr,
					1, (uint8_t (*)[6])best_fh->bss->bssid,
					STEER_MODE_BTM_REQ); /* mandate */
			if (ret)
				warn("%s: Failed to send cmdu for steering sta!\n", __func__);
		}
	}
}
#endif

struct wifi_apsta_steer_history *sta_lookup_steer_attempt(
		struct sta *s, uint8_t *src, uint8_t *dst)
{
	int i;
	uint8_t num_attempts = s->de_sta->mapsta.num_steer_hist;
	int size = (num_attempts < MAX_STEER_HISTORY) ?
				num_attempts : MAX_STEER_HISTORY;
	struct wifi_multiap_sta *mapsta;

	trace("%s: --->\n", __func__);

	mapsta = &s->de_sta->mapsta;

	/* Target BSSID is empty (error): use most recent unanswered attempt */
	if (!dst) {
		for (i = size - 1; i >= 0 ; i--) {
			if (!mapsta->steer_history[i].duration &&
					!memcmp(mapsta->steer_history[i].src_bssid, src, 6)) {
				dbg("|%s:%d| empty dst bssid - updated first unused attempt!\n",
				    __func__, __LINE__);
				return &mapsta->steer_history[i];
			}
		}
		return NULL;
	}

	dbg("|%s:%d| src="MACFMT", dst="MACFMT"\n",
	    __func__, __LINE__, MAC2STR(src), MAC2STR(dst));

	/* Target BSSID non-empty - lookup attempt based on that */
	for (i = 0; i < size; i++) {
		if (!mapsta->steer_history[i].duration &&
				!memcmp(mapsta->steer_history[i].src_bssid, src, 6) &&
				!memcmp(mapsta->steer_history[i].dst_bssid, dst, 6))
			return &mapsta->steer_history[i];
	}

	dbg("|%s:%d| dst "MACFMT" not found on attempt list!\n",
		__func__, __LINE__, MAC2STR(dst));

	/* Target BSSID not found. Check if provided in an attempt */
	for (i = 0; i < size; i++) {
		if (!mapsta->steer_history[i].duration &&
		    !memcmp(mapsta->steer_history[i].src_bssid, src, 6))
			return &mapsta->steer_history[i];
	}

	dbg("|%s:%d| auto bssid not found on attempt list!\n",
		__func__, __LINE__);

	return NULL;
}

void cntlr_update_sta_steer_counters(struct controller *c,
				     uint8_t *sta_mac,
				     uint8_t *src_bssid,
				     uint8_t *dst_bssid,
				     uint32_t mode,
				     enum steer_trigger trigger,
				     uint8_t dst_rcpi)
{
	trace("%s:--->\n", __func__);
	struct sta *s = cntlr_find_sta(c->sta_table, sta_mac);
	struct wifi_apsta_steer_history *a;
	struct wifi_multiap_sta *mapsta;

	if (!s) {
		dbg("|%s:%d| Unrecognized STA "MACFMT", skip!\n",
		    __func__, __LINE__, MAC2STR(sta_mac));
		return;
	}

	mapsta = &s->de_sta->mapsta;
	if (mapsta->num_steer_hist >= MAX_STEER_HISTORY) {
		int i;

		for (i = 0; i < MAX_STEER_HISTORY - 1; i++) {
			memcpy(&mapsta->steer_history[i], &mapsta->steer_history[i+1],
					sizeof(struct wifi_apsta_steer_history));
		}
		a = &mapsta->steer_history[MAX_STEER_HISTORY - 1];
	} else {
		a = &mapsta->steer_history[mapsta->num_steer_hist];
	}

	/* Update SteeringHistory */
	timestamp_update(&a->time);
	time(&a->steer_time);

	if (src_bssid)
		memcpy(a->src_bssid, src_bssid, 6);

	if (dst_bssid)
		memcpy(a->dst_bssid, dst_bssid, 6);

	a->trigger = trigger;
	if (trigger == STEER_TRIGGER_LINK_QUALITY) {
		a->src_rcpi = s->de_sta->rcpi;
		a->dst_rcpi = dst_rcpi;
	}

	switch (mode) {
	case STEER_MODE_ASSOC_CTL:
		a->method = STEER_METHOD_ASSOC_CTL;
		break;
	case STEER_MODE_BTM_REQ:
		a->method = STEER_METHOD_BTM_REQ;
		/* Update SteeringSummaryStats - per STA & per Network */
		s->de_sta->mapsta.stats.btm_attempt_cnt++;
		c->dlem.network.steer_summary.btm_attempt_cnt++;
		break;
	case STEER_MODE_OPPORTUNITY:
		a->method = STEER_METHOD_ASYNC_BTM;
		/*TODO: add counter for opportunity (incl blacklis count) */
		break;
	default:
		a->method = STEER_METHOD_UNKNOWN;
		break;
	}
	a->duration = 0;

	/* Record tsp for most recent steer attempt */
	timestamp_update(&s->de_sta->mapsta.stats.last_attempt_tsp);
	mapsta->num_steer_hist += 1;
}
void cntlr_notify_client_steer_req_evt(struct controller *c,
			uint8_t *bssid, uint32_t sta_nr, uint8_t stas[][6],
			uint32_t bssid_nr, uint8_t target_bssid[][6])
{
	char ev_data[1024] = {0};

	snprintf(ev_data, sizeof(ev_data),
			"{\"bssid\":\""MACFMT"\"",
			MAC2STR(bssid));

	if (sta_nr) {
		char mac[64] = {0};

		strncat(ev_data, ",\"sta_mac\":", sizeof(ev_data) - strlen(ev_data));
		snprintf(mac, sizeof(mac), "\""MACFMT"\"", MAC2STR(stas[0]));
		strncat(ev_data, mac, sizeof(ev_data) - strlen(ev_data));

		// TODO: use blob_buf directly to provide further STA MACs
	}

	if (bssid_nr) {
		char mac[64] = {0};

		strncat(ev_data, ",\"target_bssid\":", sizeof(ev_data) - strlen(ev_data));
		snprintf(mac, sizeof(mac), "\""MACFMT"\"", MAC2STR(target_bssid[0]));
		strncat(ev_data, mac, sizeof(ev_data) - strlen(ev_data));

		// TODO: use blob_buf directly to provide further target MACs
	}

	strncat(ev_data, "}", sizeof(ev_data) - strlen(ev_data));

	cntlr_notify_event(c, "client_steer_request", ev_data);
}

void cntlr_notify_client_steer_result(struct controller *c,
		uint8_t *sta_mac, int result)
{
	char ev_data[1024] = {0};

	snprintf(ev_data, sizeof(ev_data),
		 "{\"sta_mac\":\""MACFMT"\""
		 ",\"status\":%d}",
		 MAC2STR(sta_mac), result);

	cntlr_notify_event(c, "client_steer_result", ev_data);
}

void cntlr_btm_req_timer_cb(atimer_t *t)
{
	trace("%s:--->\n", __func__);

	struct sta *s = container_of(t, struct sta, btm_req_timer);
	struct node *n = s->fh->agent;
	struct controller *c = n->cntlr;

	if (s->de_sta->mapsta.pending_btm_resp_num > 0) {
		s->de_sta->mapsta.failed_steer_attempts +=
				s->de_sta->mapsta.pending_btm_resp_num;
		s->de_sta->mapsta.stats.btm_failure_cnt +=
				s->de_sta->mapsta.pending_btm_resp_num;
		c->dlem.network.steer_summary.btm_failure_cnt +=
				s->de_sta->mapsta.pending_btm_resp_num;
		s->de_sta->mapsta.pending_btm_resp_num = 0;

		cntlr_notify_client_steer_result(c, s->de_sta->macaddr,
				STEER_RESULT_FAIL_TIMEOUT);
	}
}

int cntlr_steer_sta(struct controller *c, struct sta *s,
		    struct wifi_sta_meas_report *to, uint32_t mode,
		    uint32_t reason)
{
	int ret = 0;
	uint16_t mid;

	trace("%s:--->\n", __func__);

	if (!to || hwaddr_is_zero(to->bssid)) {
		dbg("%s: steer verdict = OK, but target AP = NULL!\n", __func__);
		return 0;
	}

	if (!memcmp(to->bssid, s->bssid, 6)) {
		s->de_sta->mapsta.stats.no_candidate_cnt++;
		c->dlem.network.steer_summary.no_candidate_cnt++;
		dbg("%s: " MACFMT " connected to best AP! No steer needed.\n",
		    __func__, MAC2STR(s->de_sta->macaddr));
		return 0;
	}

	warn("%s: Try to steer " MACFMT " from " MACFMT " to " MACFMT "\n",
	     __func__, MAC2STR(s->de_sta->macaddr), MAC2STR(s->bssid), MAC2STR(to->bssid));

	UNUSED(reason);

	switch (mode) {
	case STEER_MODE_ASSOC_CTL:
		ret = cntlr_send_client_assoc_ctrl_request(c,
							   s->fh->agent->almacaddr,
							   s->bssid,
							   0,
							   10, /* block bssid for 10 sec */
							   1,
							   s->de_sta->macaddr,
							   &mid);
		if (ret) {
			warn("%s: Failed to send cmdu for assoc control!\n", __func__);
			//s->de_sta->mapsta.failed_steer_attempts++;
			return ret;
		}
		/* Keep mid & check assoc control succesful in ACK msg */
		s->latest_assoc_cntrl_mid = mid;
		dbg("%s: cmdu->cdata->hdr.mid %u\n", __func__, mid);
		break;
	case STEER_MODE_BTM_REQ:
	case STEER_MODE_OPPORTUNITY:
		ret = cntlr_send_client_steer_request(c, s->fh->agent->almacaddr,
					s->bssid, 0,
					1, (uint8_t (*)[6])s->de_sta->macaddr,
					1, (uint8_t (*)[6])to->bssid,
					mode);
		if (ret) {
			warn("%s: Failed to send cmdu for steering sta!\n", __func__);
			return ret;
		}

		/* Expect btm-resp from STA forwarded to cntlr */
		s->de_sta->mapsta.pending_btm_resp_num++;
		timer_set(&s->btm_req_timer, BTM_RESP_EXP_TIMEOUT * 1000);

		break;
	case STEER_MODE_UNDEFINED:
	default:
		dbg("%s: steer mode is undefined\n", __func__);
		return 0;
	}

	cntlr_update_sta_steer_counters(c, s->de_sta->macaddr, s->bssid, to->bssid,
				mode, STEER_TRIGGER_LINK_QUALITY, to->rcpi);

	return 0;
}

void cntlr_inform_bsteer_modules(struct controller *c, struct sta *s, uint16_t rxcmdu_type)
{
	//TODO: bSTA steer plugins
}

void cntlr_inform_steer_modules(struct controller *c, struct sta *s, uint16_t rxcmdu_type)
{
	struct steer_control *sc = NULL;
	struct steer_sta candidate = {
		.sta = s->de_sta,
		.nbrlist = NULL,
		.meas_reportlist = &s->de_sta->meas_reportlist,
		.unassoc_metriclist = &s->de_sta->umetriclist,
		.best = NULL,
	};
#if 0	// TODO
	struct node *n = NULL;
#endif
	struct netif_radio *r;

	trace("%s:--->\n", __func__);
	memcpy(candidate.bss.bssid, s->bssid, 6);
	candidate.bss.opclass = s->fh->radio->radio_el->cur_opclass.opclass[0].id; //FIXME

	r = cntlr_find_radio_with_bssid(c, s->bssid);
	if (r) {
		struct radio_policy *rp;

		rp = cntlr_get_radio_policy(&c->cfg, r->radio_el->macaddr);
		if (rp) {
			candidate.bss.rcpi_threshold = rp->rcpi_threshold;
			candidate.bss.report_rcpi_threshold = rp->report_rcpi_threshold;
			candidate.bss.report_rcpi_hysteresis_margin = rp->report_rcpi_hysteresis_margin;
			candidate.bss.util_threshold = rp->util_threshold;
			candidate.bss.report_util_threshold = rp->report_util_threshold;
		}
	}

#if 0	//TODO
	//TODO: candidate->nbrlist = cntlr->fbss_nbrlist
	// Also, move nbrlist creation to cntlr main

	INIT_LIST_HEAD(&candidate.nbrlist);
	list_for_each_entry(n, &c->nodelist, list) {
		struct netif_radio *r = NULL;
		int ret;

		list_for_each_entry(r, &n->radiolist, list) {
			struct netif_iface *p = NULL;
			uint8_t opclass = 0;
			uint8_t ctrl_channel = 0;	/* 20MHz */

			/* if (s->fh->band != r->radio_el->band) {
				continue;
			}
			*/

			ret = cntlr_radio_get_beacon_channel(r->radio_el, &opclass, &ctrl_channel);
			if (ret)
				continue;

			list_for_each_entry(p, &r->iflist, list) {
				struct steer_sta_target_bss *t;

				if (!p->bss || !s->fh || !s->fh->bss)
					continue;

				if (memcmp(p->bss->ssid, s->fh->bss->ssid, 33))
					continue;

				t = calloc(1, sizeof(*t));
				if (t) {
					memcpy(t->bssid, p->bss->bssid, 6);
					t->opclass = opclass;
					t->channel = ctrl_channel;
					list_add_tail(&t->list, &candidate.nbrlist);
				}
			}
		}
	}
#endif

	list_for_each_entry(sc, &c->sclist, list) {
		cntlr_maybe_steer_sta(sc, &candidate, rxcmdu_type);

		switch (candidate.verdict) {
		case STEER_VERDICT_OK:
			if (c->cfg.steer.plugin_policy == STEER_PLUGIN_POLICY_OR) {
				cntlr_warn(LOG_STEER,
					   "Steer STA: " MACFMT ", src-BSSID:" MACFMT
					   ", target-BSSID:" MACFMT ", reason: %s\n",
					   MAC2STR(s->de_sta->macaddr),
					   MAC2STR(s->bssid),
					   MAC2STR(candidate.best->bssid),
					   candidate.reason == STEER_REASON_LOW_RCPI ? "link quality" :
					   candidate.reason == STEER_REASON_LOW_THPUT ? "phyrate" :
					   "unknown");

				cntlr_steer_sta(c, s, candidate.best,
						STEER_MODE_BTM_REQ,
						candidate.reason);

				return;
			}

			break;
		case STEER_VERDICT_NOK:
			if (c->cfg.steer.plugin_policy == STEER_PLUGIN_POLICY_AND)
				return;

			break;
		default:
			break;
		}
	}
}
